/*****************************************************************************/
/*
                                 Auth.h
*/
/*****************************************************************************/

#ifndef AUTH_H_LOADED
#define AUTH_H_LOADED 1

#include "wasd.h"

/* some storage maximums */
#define  AUTH_MAX_DIGEST_RESPONSE      32
#define  AUTH_MAX_PASSWORD_LENGTH      15
#define  AUTH_MAX_REALM_GROUP_LENGTH   32
#define  AUTH_MAX_USERNAME_LENGTH      15
#define  AUTH_MAX_FULLNAME_LENGTH      31
#define  AUTH_MAX_EMAIL_LENGTH        127
#define  AUTH_MAX_CONTACT_LENGTH      127

/* password minimum (when changing in HTADMIN.C) */
#define  AUTH_MIN_PASSWORD 6

/* maximum number of VMS identifiers in array */
#define  AUTH_MAX_VMS_IDENTIFIERS 127

/* constants used to identify authorization schemes */
#define  AUTH_SCHEME_BASIC   0x00000001
#define  AUTH_SCHEME_DIGEST  0x00000002

/* constants for bits for flags in HTA database */
#define  AUTH_FLAG_GET         0x00000001
#define  AUTH_FLAG_HEAD        0x00000002
#define  AUTH_FLAG_POST        0x00000004
#define  AUTH_FLAG_PUT         0x00000008
#define  AUTH_FLAG_DELETE      0x00000010
#define  AUTH_FLAG_TRACE       0x00000020
#define  AUTH_FLAG_HTTPS_ONLY  0x20000000
#define  AUTH_FLAG_MANAGER     0x40000000
#define  AUTH_FLAG_ENABLED     0x80000000

#define  AUTH_READWRITE_ACCESS  HTTP_METHOD_DELETE | \
                                HTTP_METHOD_GET | \
                                HTTP_METHOD_HEAD | \
                                HTTP_METHOD_POST | \
                                HTTP_METHOD_PUT

#define  AUTH_READONLY_ACCESS  HTTP_METHOD_GET | \
                               HTTP_METHOD_HEAD

#define  AUTH_REALM_EXTERNAL     "EXTERNAL"
#define  AUTH_REALM_FAIL         "-FAIL-"
#define  AUTH_REALM_NONE         "NONE"
#define  AUTH_REALM_PROMISCUOUS  "PROMISCUOUS"
#define  AUTH_REALM_VMS          "VMS"
#define  AUTH_REALM_WORLD        "WORLD"

#define  AUTH_SOURCE_FAIL         1
#define  AUTH_SOURCE_NONE         2
#define  AUTH_SOURCE_WORLD        3
#define  AUTH_SOURCE_EXTERNAL     4
#define  AUTH_SOURCE_VMS          5
#define  AUTH_SOURCE_LIST         6
#define  AUTH_SOURCE_HTA          7
#define  AUTH_SOURCE_ID           8
#define  AUTH_SOURCE_WASD_ID      9
#define  AUTH_SOURCE_PROMISCUOUS 10

#define  HTA_DIRECTORY     "HT_AUTH:"
#define  HTA_FILE_TYPE     ".HTA"
#define  HTL_FILE_TYPE     ".HTL"

/* bogus VMS status codes, these can be passed back and detected */
#define  AUTH_RESTRICTED_NOT          0x00000001
#define  AUTH_LOGOUT                  0x0fffff1a
#define  AUTH_RESTRICTED_BY_HOSTNAME  0x0fffff2a
#define  AUTH_RESTRICTED_BY_PROTOCOL  0x0fffff3a
#define  AUTH_RESTRICTED_BY_USERNAME  0x0fffff4a
#define  AUTH_RESTRICTED_BY_HOUR      0x0fffff5a
#define  AUTH_RESTRICTED_BY_OTHER     0x0fffff6a
#define  AUTH_RESTRICTED_BY_FAIL      0x0fffff7a

#define  AUTH_RESTRICTED_FAIL "-FAIL-"

/* identifier names controlling SYSUAF authentication with /SYSUAF=ID */
#define  AUTH_HTTPS_ONLY_VMS_ID   "WASD_HTTPS_ONLY"

/* identifier names controlling SYSUAF authentication with /SYSUAF=WASD */
#define  AUTH_WASD__GROUP_VMS_ID  "WASD_VMS__"
#define  AUTH_WASD_HTTPS_VMS_ID   "WASD_VMS_HTTPS"
#define  AUTH_WASD_PWD_VMS_ID     "WASD_VMS_PWD"
#define  AUTH_WASD_WRITE_VMS_ID   "WASD_VMS_RW"
#define  AUTH_WASD_READ_VMS_ID    "WASD_VMS_R"

/* when formatting can string */
#define  AUTH_CAN_FORMAT_SHORT  0
#define  AUTH_CAN_FORMAT_HTML   1
#define  AUTH_CAN_FORMAT_CGI    2

/**************/
/* structures */
/**************/

/*
   Store these structures naturally-aligned on AXP.
   Uses a bit more storage but makes access as efficient as possible.
*/

#ifdef __ALPHA
#   pragma member_alignment __save
#   pragma member_alignment
#endif

#define AuthCurrentDatabaseVersion 0x01

/* record structure of on-disk HTA database */
struct AuthHtRecordStruct
{
   char  UserName [AUTH_MAX_USERNAME_LENGTH+1];
   char  FullName [AUTH_MAX_FULLNAME_LENGTH+1];
   char  Email [AUTH_MAX_EMAIL_LENGTH+1];
   int  UserNameLength;
   unsigned long  Flags;
   unsigned long  HashedPwd [2];
   unsigned char  A1DigestLoCase [16];
   unsigned char  A1DigestUpCase [16];
   int  ChangeCount;
   unsigned long  LastChangeBinTime [2];
   int  AccessCount;
   unsigned long  LastAccessBinTime [2];
   int  FailureCount;
   unsigned long  LastFailureBinTime [2];
   unsigned long  AddedBinTime [2];
   unsigned char  Contact [AUTH_MAX_CONTACT_LENGTH+1];
   unsigned char  FillTo512Bytes [115];
   unsigned char  DatabaseVersion;
};

/* in-memory path authorization structure */
struct AuthPathRecordStruct
{
   struct AuthPathRecordStruct  *PrevPtr;
   struct AuthPathRecordStruct  *NextPtr;

   int  GroupReadLength,   
        GroupRestrictListLength,
        GroupWriteLength,   
        PathLength,
        RealmLength,
        RealmDescriptionLength,
        WorldRestrictListLength,
        VirtualServerHostPortLength;

   unsigned long  AuthGroupCan,
                  AuthWorldCan,
                  GroupReadVmsIdentifier,
                  GroupWriteVmsIdentifier,
                  RealmVmsIdentifier,
                  SourceRealm,
                  SourceGroupRead,
                  SourceGroupWrite,
                  VirtualServerHostNameHashValue;

   char  *GroupReadPtr,
         *GroupWritePtr,
         *PathPtr,
         *RealmPtr,
         *RealmAccessPtr,
         *RealmDescriptionPtr,
         *GroupRestrictListPtr,
         *VirtualServerHostPortPtr,
         *WorldRestrictListPtr;

   /* string data is stored here and pointed to by above pointers */
};

/* in-memory user authentication cache structure */
struct AuthCacheRecordStruct
{
   /* this is the structure used by the VMS binary tree library functions */
   struct AuthCacheRecordStruct  *LeftLink;
   struct AuthCacheRecordStruct  *RightLink;
   unsigned short  Reserved;

   boolean  CanChangeSysUafPwd,
            HttpsOnly,
            SysUafAuthenticated;

   int  AccessCount,
        AuthUserCan,
        BasicCount,
        CacheMinutes,
        DataBaseCount,
        DigestCount,
        FailureCount,
        GroupReadLength,
        GroupWriteLength,
        RealmLength,
        RealmDescriptionLength,
        RevalidateUserMinutes,
        UserNameLength;

   unsigned short  VmsUserProfileLength;

   unsigned long  SourceRealm,
                  SourceGroupRead,
                  SourceGroupWrite,
                  UaiPrimeDays,
                  UaiNetworkAccessP,
                  UaiNetworkAccessS,
                  UaiRemoteAccessP,
                  UaiRemoteAccessS;

   unsigned long  LastAccessBinTime [2];

   unsigned char  *VmsUserProfilePtr;

   char  Realm [AUTH_MAX_REALM_GROUP_LENGTH+1],
         RealmDescription [AUTH_MAX_REALM_GROUP_LENGTH+1],
         GroupWrite [AUTH_MAX_REALM_GROUP_LENGTH+1],
         GroupRead [AUTH_MAX_REALM_GROUP_LENGTH+1],
         UserName [AUTH_MAX_USERNAME_LENGTH+1],
         Password [AUTH_MAX_PASSWORD_LENGTH+1],
         DigestResponse [AUTH_MAX_DIGEST_RESPONSE+1];
};

/* structure used to hold an authorization database loaded from file */
struct AuthConfigStruct
{
   int  LineNumber,
        ProblemCount,
        ProblemReportLength;

   unsigned long  LoadBinTime [2],
                  RevBinTime [2];

   char  *LinePtr,
         *ProblemReportPtr;

   char  LoadFileName [256];

   struct AuthPathRecordStruct  *PathListHead,
                                *PathListTail;

   /* only used during HTTP request */
   struct RequestStruct  *RequestPtr;
};

#ifdef __ALPHA
#   pragma member_alignment __restore
#endif

/***********************/
/* function prototypes */
/***********************/

AuthInit();
int Authorize (struct RequestStruct*);
int AuthParseAuthorization (struct RequestStruct*);
int AuthVerifyPassword (struct RequestStruct*);
int AuthVerifyVmsPassword (struct RequestStruct*,
                           struct AuthCacheRecordStruct*);
int AuthReadHtDatabase (struct RequestStruct*, char*, boolean);
int AuthReadSimpleList (struct RequestStruct*, char*, boolean);

AuthUserResetRecord (char*, char*, char*);
AuthCacheTreeFree ();
int AuthCacheTreeAllocateRecord (struct AuthCacheRecordStruct*,
                                 struct AuthCacheRecordStruct**,
                                 unsigned long);
int AuthCacheTreeCompareRecord (struct AuthCacheRecordStruct*,
                                struct AuthCacheRecordStruct*,
                                unsigned long);
int AuthCacheTreeFreeRecord (struct AuthCacheRecordStruct*, unsigned long);
AuthCacheTreeReport (struct RequestStruct*, void*);
int AuthCacheTreeReportRecord (struct AuthCacheRecordStruct*,
                               struct RequestStruct*);
int AuthCacheTreeEntryReset (struct AuthCacheRecordStruct*,
                             struct AuthCacheRecordStruct*);

char* AuthCanString (unsigned long, int);

int AuthCheckVmsUserAccess (struct RequestStruct*, char*, int);
int AuthCheckWriteAccess (struct RequestStruct*, char*, int);
int AuthCreateVmsUserProfile (struct RequestStruct*);

int AuthConfigIdentifier (char*, unsigned long*);
AuthConfigReport (struct AuthConfigStruct*, char*, ...);
int AuthControlReload (struct RequestStruct*, void*);
int AuthPathAdd (struct AuthConfigStruct*, char*,
                 char*, char*, unsigned long, unsigned long,
                 char*, unsigned long, unsigned long,
                 char*, unsigned long, unsigned long,
                 char*, char*, char*, 
                 unsigned long, unsigned long);
int AuthPathCheck (struct RequestStruct*, struct AuthConfigStruct*);
int AuthPathConfig (struct AuthConfigStruct*);
int AuthPathListFree (struct AuthConfigStruct*);
int AuthPathReport (struct RequestStruct*, void*, boolean, char*);
int AuthPathReportNow (struct RequestStruct*, boolean,
                       struct AuthConfigStruct*, char*);
int AuthHoldsVmsIdentifier (struct RequestStruct*, char*, unsigned long);
int AuthLoadVmsIdentifiers (struct RequestStruct*, unsigned long);
int AuthSysUafCheckIdentifier (struct RequestStruct*);
char* AuthSourceString (char*, unsigned long);

#endif /* AUTH_H_LOADED */

/*****************************************************************************/
