/*****************************************************************************/
/*
                                 SeSoLa.c


    THE GNU GENERAL PUBLIC LICENSE APPLIES DOUBLY TO ANYTHING TO DO WITH
                    AUTHENTICATION AND AUTHORIZATION!

    This package is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License, or any later
    version.

>   This package is distributed in the hope that it will be useful,
>   but WITHOUT ANY WARRANTY; without even the implied warranty of
>   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
>   GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


Named "SeSoLa" to avoid any confusion and/or conflict with OpenSSL routines.

SSL I/O is implemented as a OpenSSL BIO_METHOD, named "sesola_method". It
provides NON-BLOCKING SSL input/output. All routines that are part of this
functionality are named "sesola_..." and are grouped towards the end of this
module.


OPENSSL
-------
With the retirement of Eric Young from OpenSource software and the creation of
the OpenSSL organisation this module will be based on this group's packages and
in particular the the VMS port supported by Richard Levitte.

  richard@levitte.org
  http://www.free.lp.se/openssl/
  http://www.openssl.org/

The initial adaption has been made using v0.9.3.


SSL COMMAND LINE OPTIONS
------------------------
The /SSL= qualifier allows certain SSL runtime parameters to be set.  The
default is support of SSLv2/SSLv3, certificate file provided by HTTPD$SSL_CERT,
all ciphers, and a session cache of 128.

  /SSL=2                  support only SSLv2 (pre-v6.0.0)
  /SSL=3                  support only SSLv3 (pre-v6.0.0)
  /SSL=23                 support both SSLv2 and SSLv3 (pre-v6.0.0)
  /SSL=SSLv2              support only SSLv2
  /SSL=noSSLv2            turn SSLv2 option off
  /SSL=SSLv3              support only SSLv3
  /SSL=noSSLv3            turn SSLv3 option off
  /SSL=(SSLv2,SSLv3)      support both SSLv2 and SSLv3
  /SSL=TLSv1              support only TLSv1 (etc.)
  /SSL=noTLSv1            turn TLSv1 option off
  /SSL=(OPTIONS=0xnn)     SSL_OP_.. options (see [.INCLUDE.OPENSSL]SSL.H)
  /SSL=(CERT=file)        location of default OpenSSL-PEM certificate file
  /SSL=(CIPHER=list)      semi-colon-separated list of supported ciphers
  /SSL=(CACHE=integer)    size of session cache
  /SSL=(KEY=file)         location of default OpenSSL-PEM private key file
                          (if separate from the certificate information)

Multiple parameters may be included by separating with commas.  Example:

  /SSL=(CERT=HT_ROOT:[LOCAL]SITE.PEM,SSLV2,SSLV3)


LOGICAL/SYMBOL NAMES
--------------------
The following logical or symbol names can be used to provide runtime
information to the SSL functionality.  Configuration options provided by /SSL=
and [service] or /SERVICE= parameters override anything provided via these
logical/symbol names.

  HTTPD$SSL_CERT          location of default OpenSSL-PEM certificate file
  HTTPD$SSL_CIPHER        comma-separated list of default supported ciphers
  HTTPD$SSL_KEY           location of default OpenSSL-PEM private key file
  HTTPD$SSL_PARAMS        the equivalent functionality of the /SSL= qualifier


VERSION HISTORY
---------------
11-SEP-99  MGD  tested against OpenSSL v0.9.4
18-AUG-99  MGD  add certificate CA and DN to SSL report,
                bugfix; certificate/key pairs when initializing service
12-JUN-99  MGD  refine SSL connection handling
26-MAY-99  MGD  allow some SSL options to be set if desired,
                set SSL_OP_NON_EXPORT_FIRST (for some certificate processing)
18-APR-99  MGD  improve error reporting during certificate loading
03-APR-99  MGD  support OpenSSL 0.9.3
                (with initial, backward support for SSLeay 0.8 & 0.9), 
                add WATCH callbacks
31-MAR-99  MGD  bugfix; report request from non-SSL port
20-JAN-99  MGD  report format refinement
07-NOV-98  MGD  WATCH facility
24-OCT-98  MGD  per-service context (implies per-service certificate)
01-JUN-98  MGD  builds OK with SSLeay 0.9.0b
25-JAN-98  MGD  initial development for v5.0, SSLeay v0.8.1
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* VMS related header files */
#include <ssdef.h>
#include <stsdef.h>

/* application header files */
#include "wasd.h"
#include "httpd.h"

#define WASD_MODULE "SESOLA"

/***************************************/
#ifdef SESOLA  /* secure sockets layer */
/***************************************/

/* OpenSSL header files (0.9.3ff) */
#ifdef SESOLA_OPENSSL
#   include "openssl/bio.h"
#   include "openssl/buffer.h"
#   include "openssl/crypto.h"
#   include "openssl/err.h"
#   include "openssl/ssl.h"
#   include "openssl/X509.h"
# else
/*  WASD SSLeay 0.8.1, 0.9.0 */
#   include "bio.h"
#   include "buffer.h"
#   include "crypto.h"
#   include "err.h"
#   include "ssl.h"
#   include "X509.h"
#endif

/**********/
/* macros */
/**********/

#define DEFAULT_SESSION_CACHE_SIZE 128

#define SESOLA_UNDEFINED          0
#define SESOLA_ACCEPT             1
#define SESOLA_SHUTDOWN           2
#define SESOLA_DATA_BLOCKING      3
#define SESOLA_DATA_NON_BLOCKING  4

#define SESOLA_PROTOCOL_SSLV2  0x01
#define SESOLA_PROTOCOL_SSLV3  0x02
#define SESOLA_PROTOCOL_TLSV1  0x04

/**************************************/
/* macros to support SSLeay 0.8 & 0.9 */
/**************************************/

#ifndef OPENSSL_VERSION_NUMBER

#define SSL_library_init SSLeay_add_ssl_algorithms

#endif

/*******************/
/* data structures */
/*******************/

struct SeSoLaStruct
{
   /* OpenSSL structures */
   SSL  *SslPtr;
   BIO  *BioPtr;
   BIO  *BioSslPtr;
   SSL_CTX  *SeSoLaCtx;

   /* pointer back to the request owning this structure */
   struct RequestStruct  *rqptr;

   /*
       These "in-progress" flags are to provide an additional indication of
       when a OpenSSL routine has used non-blocking I/O.  Onnly valid within
       the single AST-delivery context (i.e. to be checked immediately after
       using an OpenSSL function, before returning out of the current AST
       delivery context), because as soon as the I/O AST is delivered the
       approriate flags will be reset.
   */
   boolean  ReadInProgress,
            ReadCompleted,
            WriteInProgress,
            WriteCompleted;

   int  /* */
        AcceptCount,
        /* size of read data buffer (for AST routine) */
        ReadDataSize,
        /* flag indicating purpose and/or type of network read */
        ReadType,
        /* number of characters to be written to network */
        WriteDataLength,
        /* flag indicating purpose and/or type of network write */
        WriteType;

  char  /* during SSL connect accept points to start of buffer */
        *AcceptReadDataPtr,
        /* pointer to read data buffer (for AST routine use) */
        *ReadDataPtr,
        /* pointer to write data buffer (for AST routine use) */
        *WriteDataPtr;

   /* SSL post-processing, stores the pointer to the AST routine */
   void  *ReadAstFunctionPtr,
         *WriteAstFunctionPtr;
};

/******************/
/* global storage */
/******************/

boolean  ProtocolHttpsAvailable,
         ProtocolHttpsConfigured,
         SeSoLaOptions;

int  SeSoLaSessionCacheSize,
     SeSoLaOptionsOff,
     SeSoLaProtocolVersion;

char  *SeSoLaDefaultCertPtr,
      *SeSoLaDefaultCipherListPtr,
      *SeSoLaDefaultKeyPtr,
      *SeSoLaProtocolsPtr;

char  SeSoLaParams [256] = "";

char  ErrorSslPort [] = "This is an SSL (&quot;https:&quot;) port.",
      ErrorSslReportService [] =
        "This report is only available via an SSL service!",
      HttpdSesola [] = " SSL";

BIO_METHOD  *SesolaBioMemPtr;


/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern int  AdminWatchEnabled;
extern int  NetReadBufferSize;
extern int  OutputBufferSize;
extern char  ErrorSanityCheck[];
extern char  ServerHostPort[];
extern char  Utility[];
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;
extern struct MsgStruct  Msgs;
extern struct ServiceStruct  *ServiceListHead;

/**********************************/
/* BIO method function prototypes */
/**********************************/

int sesola_write (BIO*, char*, int);
int sesola_write_ast (struct RequestStruct*);
int sesola_read (BIO*, char*, int);
int sesola_read_ast (struct RequestStruct*);
int sesola_puts (BIO*, char*);
int sesola_gets (BIO*, char*, int);
long sesola_ctrl (BIO*, int, long, char*);
int sesola_new (BIO*);
int sesola_free (BIO*);

int SesolaWatchBioCallback (BIO*, int, char*, int, long, long);
SesolaWatchInfoCallback (SSL*, int, int);

BIO_METHOD *BIO_s_sesola();

BIO_METHOD sesola_method =
{
   BIO_TYPE_FD,
   "WASD SESOLA",
   sesola_write,
   sesola_read,
   sesola_puts,
   sesola_gets,
   sesola_ctrl,
   sesola_new,
   sesola_free,
};

/*****************************************************************************/
/*
*/

SeSoLaInit ()

{
   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaInit()\n");

   fprintf (stdout, "%%%s-I-SSL, %s\n",
            Utility, SSLeay_version(SSLEAY_VERSION));

   /* command-line parameters, if none then use HTTPD$SSL_PARAMS */
   cptr = SeSoLaParams;
   if (!*cptr) cptr = getenv ("HTTPD$SSL_PARAMS");
   if (cptr == NULL) cptr = "";

   if (strsame (cptr, "/NOSSL", -1))
   {
      fprintf (stdout, "%%%s-W-SSL, disabled via /NOSSL\n", Utility);
      return;
   }

   while (*cptr)
   {
      while (*cptr && (*cptr == ',' || *cptr == ')')) cptr++;
      sptr = cptr;
      while (*cptr && *cptr != ',' && *cptr != ')') cptr++;
      if (*cptr) *cptr++ = '\0';
      if (!*sptr) break;
      if (Debug) fprintf (stdout, "sptr |%s|\n", sptr); 

      /* the 'digits' are for backward compatibility with pre-v6.0.0 */
      if (strsame (sptr, "2", -1))
         SeSoLaProtocolVersion |= SESOLA_PROTOCOL_SSLV2;
      else
      if (strsame (sptr, "3", -1))
         SeSoLaProtocolVersion |= SESOLA_PROTOCOL_SSLV3;
      else
      if (strsame (sptr, "23", -1))
         SeSoLaProtocolVersion |= SESOLA_PROTOCOL_SSLV2 |
                                  SESOLA_PROTOCOL_SSLV3;
      else
      if (strsame (sptr, "SSLv2", -1))
         SeSoLaProtocolVersion |= SESOLA_PROTOCOL_SSLV2;
      else
      if (strsame (sptr, "noSSLv2", -1))
         SeSoLaOptionsOff |= SSL_OP_NO_SSLv2;
      else
      if (strsame (sptr, "SSLv3", -1))
         SeSoLaProtocolVersion |= SESOLA_PROTOCOL_SSLV3;
      else
      if (strsame (sptr, "noSSLv3", -1))
         SeSoLaOptionsOff |= SSL_OP_NO_SSLv3;
      else
      if (strsame (sptr, "TLSv1", -1))
         SeSoLaProtocolVersion |= SESOLA_PROTOCOL_TLSV1;
      else
      if (strsame (sptr, "noTLSv1", -1))
         SeSoLaOptionsOff |= SSL_OP_NO_TLSv1;
      else
      if (strsame (sptr, "CERT=", 5))
         SeSoLaDefaultCertPtr = sptr + 5;
      else
      if (strsame (sptr, "CIPHER=", 7))
         SeSoLaDefaultCipherListPtr = sptr + 7;
      else
      if (strsame (sptr, "CACHE=", 6))
         SeSoLaSessionCacheSize = atoi(sptr+6);
      else
      if (strsame (sptr, "KEY=", 4))
         SeSoLaDefaultKeyPtr = sptr + 4;
      else
      if (strsame (sptr, "OPTIONS=", 8))
      {
         if (strsame (sptr+8, "ALL", -1))
            SeSoLaOptions = -1;
         else
            SeSoLaOptions = strtol (sptr+8, NULL, 16);
      }
      else
      {
         fprintf (stdout, "%%%s-E-SSL, unknown SSL parameter\n \\%s\\\n",
                  Utility, sptr);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }
   }

   /* default is to support both SSL v2 and v3 */
   if (!(SeSoLaProtocolVersion & SESOLA_PROTOCOL_SSLV2) &&
       !(SeSoLaProtocolVersion & SESOLA_PROTOCOL_SSLV3) &&
       !(SeSoLaProtocolVersion & SESOLA_PROTOCOL_TLSV1))
      SeSoLaProtocolVersion = SESOLA_PROTOCOL_SSLV2 |
                              SESOLA_PROTOCOL_SSLV3;

   if ((SeSoLaProtocolVersion & SESOLA_PROTOCOL_SSLV2) &&
       (SeSoLaProtocolVersion & SESOLA_PROTOCOL_SSLV3))
      SeSoLaProtocolsPtr = "SSLv2/v3";
   else
   if (SeSoLaProtocolVersion & SESOLA_PROTOCOL_SSLV2)
      SeSoLaProtocolsPtr = "SSLv2";
   else
   if (SeSoLaProtocolVersion & SESOLA_PROTOCOL_SSLV3)
      SeSoLaProtocolsPtr = "SSLv3";
   else
   if (SeSoLaProtocolVersion & SESOLA_PROTOCOL_TLSV1)
      SeSoLaProtocolsPtr = "TLSv1";
   fprintf (stdout, "%%%s-I-SSL, protocol(s) %s\n",
            Utility, SeSoLaProtocolsPtr);

   /* default session cache size */
   if (!SeSoLaSessionCacheSize)
      SeSoLaSessionCacheSize = DEFAULT_SESSION_CACHE_SIZE;

   SSL_load_error_strings ();

   SSL_library_init (); 

   ProtocolHttpsAvailable = true;

   if (SeSoLaDefaultCertPtr == NULL)
      SeSoLaDefaultCertPtr = getenv ("HTTPD$SSL_CERT");

   if (SeSoLaDefaultKeyPtr == NULL)
      SeSoLaDefaultKeyPtr = getenv ("HTTPD$SSL_KEY");

   if (SeSoLaDefaultCipherListPtr == NULL)
      SeSoLaDefaultCipherListPtr = getenv ("HTTPD$SSL_CIPHER");

   SesolaBioMemPtr = BIO_new (BIO_s_mem());
   if (SesolaBioMemPtr == NULL)
   {
      fprintf (stdout, "%%%s-E-SSL, BIO_new(BIO_s_mem()) failed\n", Utility);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }
}

/*****************************************************************************/
/*
Scan through the list of services initializing an SSL context for each SSL
service.
*/

SeSoLaInitServiceList ()

{
   register char  *cptr, *sptr, *zptr;
   register struct ServiceStruct  *svptr;

   boolean  CipherListSupplied,
            KeySupplied;
   int  value;
   SSL_CTX  *SeSoLaCtx;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaInitServiceList()\n");

   for (svptr = ServiceListHead; svptr != NULL; svptr = svptr->NextPtr)
   {
      if (svptr->RequestScheme != SCHEME_HTTPS) continue;

      if (!svptr->SeSoLaCert[0])
      {
         if (SeSoLaDefaultCertPtr == NULL)
            ErrorExitVmsStatus (0, "default server certificate not specified", FI_LI);

         zptr = (sptr = svptr->SeSoLaCert) + sizeof(svptr->SeSoLaCert);
         for (cptr = SeSoLaDefaultCertPtr;
              *cptr && sptr < zptr;
              *sptr++ = toupper(*cptr++));
         if (sptr >= zptr) ErrorExitVmsStatus (0, "string overflow", FI_LI);
         *sptr = '\0';
      }

      if (!svptr->SeSoLaKey[0])
      {
         zptr = (sptr = svptr->SeSoLaKey) + sizeof(svptr->SeSoLaKey);
         if (SeSoLaDefaultKeyPtr == NULL)
         {
            /* if no default key then assume it's in the certificate file */
            KeySupplied = false;
            for (cptr = svptr->SeSoLaCert;
                 *cptr && sptr < zptr;
                 *sptr++ = toupper(*cptr++));
         }
         else
         {
            /* use the separately specified key */
            KeySupplied = true;
            for (cptr = SeSoLaDefaultKeyPtr;
                 *cptr && sptr < zptr;
                 *sptr++ = toupper(*cptr++));
         }
         if (sptr >= zptr) ErrorExitVmsStatus (0, "string overflow", FI_LI);
         *sptr = '\0';
      }
      else
         KeySupplied = true;

      if (!svptr->SeSoLaCipherList[0])
      {
         if (SeSoLaDefaultCipherListPtr == NULL)
            CipherListSupplied = false;
         else
         {
            CipherListSupplied = true;
            zptr = (sptr = svptr->SeSoLaCipherList) +
                   sizeof(svptr->SeSoLaCipherList);
            for (cptr = SeSoLaDefaultCipherListPtr;
                 *cptr && sptr < zptr;
                 *sptr++ = *cptr++);
            if (sptr >= zptr) ErrorExitVmsStatus (0, "string overflow", FI_LI);
            *sptr = '\0';
         }
      }
      else
         CipherListSupplied = true;

      fprintf (stdout,
"%%%s-I-SSL, %s\n\
-%s-I-CERT, %s\n",
               Utility, svptr->ServerHostPort,
               Utility, svptr->SeSoLaCert);

      if (KeySupplied)
         fprintf (stdout, "-%s-I-KEY, %s\n",
                  Utility, svptr->SeSoLaKey);

      if (CipherListSupplied)
         fprintf (stdout, "-%s-I-CIPHER, %s\n",
                  Utility, svptr->SeSoLaCipherList);

      if ((SeSoLaProtocolVersion & SESOLA_PROTOCOL_SSLV2) &&
          (SeSoLaProtocolVersion & SESOLA_PROTOCOL_SSLV3))
         SeSoLaCtx = SSL_CTX_new (SSLv23_server_method());
      else
      if (SeSoLaProtocolVersion & SESOLA_PROTOCOL_SSLV2)
         SeSoLaCtx = SSL_CTX_new (SSLv2_server_method());
      else
      if (SeSoLaProtocolVersion & SESOLA_PROTOCOL_SSLV3)
         SeSoLaCtx = SSL_CTX_new (SSLv3_server_method());
      else
      if (SeSoLaProtocolVersion & SESOLA_PROTOCOL_TLSV1)
         SeSoLaCtx = SSL_CTX_new (TLSv1_server_method());

      if (SeSoLaCtx == NULL)
      {
         SeSoLaPrintOpenSslErrorList ();
         ErrorExitVmsStatus (0, "SSL_CTX_new()", FI_LI);
      }

      /* turn on SYSPRV to ensure access to the certificate file */
      EnableSysPrv ();

      value = SSL_CTX_use_certificate_file (SeSoLaCtx,
                                            svptr->SeSoLaCert,
                                            SSL_FILETYPE_PEM);
      if (!value)
      {
         SeSoLaPrintOpenSslErrorList ();
         ErrorExitVmsStatus (0, "SSL_CTX_use_certificate_file()", FI_LI);
      }

      value = SSL_CTX_use_RSAPrivateKey_file (SeSoLaCtx,
                                              svptr->SeSoLaKey,
                                              SSL_FILETYPE_PEM);
      if (!value)
      {
         SeSoLaPrintOpenSslErrorList ();
         ErrorExitVmsStatus (0, "SSL_CTX_use_RSAPrivateKey_file()", FI_LI);
      }

      value = SSL_CTX_check_private_key (SeSoLaCtx);
      if (!value)
      {
         SeSoLaPrintOpenSslErrorList ();
         ErrorExitVmsStatus (0, "SSL_CTX_check_private_key()", FI_LI);
      }

      DisableSysPrv ();

      if (svptr->SeSoLaCipherList[0])
         SSL_CTX_set_cipher_list (SeSoLaCtx, svptr->SeSoLaCipherList);

      if (SeSoLaSessionCacheSize)
         SSL_CTX_sess_set_cache_size (SeSoLaCtx, SeSoLaSessionCacheSize);

      /* see [.APPS]S_SERVER.C */
      SSL_CTX_set_quiet_shutdown (SeSoLaCtx, 1);

      /* for option values, see [.INCLUDE.OPENSSL]SSL.H */
      if (SeSoLaOptions)
      {
         fprintf (stdout, "%%%s-W-SSL, options 0x%08.08x\n",
                  Utility, SeSoLaOptions);
         SSL_CTX_set_options (SeSoLaCtx, SeSoLaOptions);
      }
      else
      {
         /* see [.APPS]S_SERVER.C */
         SSL_CTX_set_options (SeSoLaCtx, SSL_OP_ALL);
         SSL_CTX_set_options (SeSoLaCtx, SeSoLaOptionsOff);
         SSL_CTX_set_options (SeSoLaCtx, SSL_OP_NON_EXPORT_FIRST);
      }

      (SSL_CTX*)svptr->SeSoLaCtx = SeSoLaCtx;
   }
}

/*****************************************************************************/
/*
Print out the OpenSSL error list.  For reporting errors associated with
certificate loading.  Based on [.CRYPTO.ERR]ERR_PRN.C.
*/

SeSoLaPrintOpenSslErrorList ()

{
   int  LineNumber,
        Flags;
   unsigned long  ErrorNumber;
   char  *FileNamePtr,
         *DataStringPtr;
   char  Buffer [200];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaPrintOpenSslErrorList()\n");

   while (ErrorNumber =
          ERR_get_error_line_data (&FileNamePtr, &LineNumber,
                                   &DataStringPtr, &Flags))
   {
      fprintf (stdout, "%%%s-E-SSL, \"%s\" (%s %d)%s%s%s\n",
               Utility,
               ERR_error_string (ErrorNumber, Buffer),
               FileNamePtr, LineNumber,
               ((Flags & ERR_TXT_STRING) ? " \"" : ""),
               ((Flags & ERR_TXT_STRING) ? DataStringPtr : ""),
               ((Flags & ERR_TXT_STRING) ? "\"" : ""));
   }
}

/*****************************************************************************/
/*
Begin an SSL transaction for a request. Create the SESOLA structure used to
store HTTPd SSL-related used during the transaction, initialize the OpenSSL
structures required, then begin the OpenSSL accept functionality.
*/

SeSoLaBegin (struct RequestStruct *rqptr)

{
   register struct SeSoLaStruct  *sesolaptr;

   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaBegin() %d\n", rqptr);

   /* note that this IS NOT HEAP memory, and must be explicitly VmFree()d */
   rqptr->SeSoLaPtr = sesolaptr =
      (struct SeSoLaStruct*)VmGet (sizeof(struct SeSoLaStruct));

   sesolaptr->SeSoLaCtx = (SSL_CTX*)rqptr->ServicePtr->SeSoLaCtx;

   /* set the SeSoLa structure request pointer */
   sesolaptr->rqptr = rqptr;

   if (Debug) fprintf (stdout, "SSL_new()\n");
   sesolaptr->SslPtr = SSL_new (sesolaptr->SeSoLaCtx);
   if (sesolaptr->SslPtr == NULL)
   {
      fprintf (stdout, "%%%s-W-SSL, SSL_new(sesolaptr->SeSoLaCtx) failed\n",
               Utility);
      SeSoLaFree (rqptr);

      if (rqptr->AdminWatchItem &&
          (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA, "BEGIN failed");

      return;
   }

   if (Debug) fprintf (stdout, "BIO_new()\n");
   sesolaptr->BioPtr = BIO_new (BIO_s_sesola());
   if (sesolaptr->BioPtr == NULL)
   {
      fprintf (stdout, "%%%s-W-SSL, BIO_new(BIO_s_sesola()) failed\n",
               Utility);
      SeSoLaFree (rqptr);

      if (rqptr->AdminWatchItem &&
          (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA, "BEGIN failed");

      return;
   }
   /* set the "BIO_s_sesola" pointer to the SESOLA structure */
   sesolaptr->BioPtr->ptr = sesolaptr;

   /* set up the SSL filter */
   if (Debug) fprintf (stdout, "BIO_new()\n");
   sesolaptr->BioSslPtr = BIO_new (BIO_f_ssl());
   if (sesolaptr->BioSslPtr == NULL)
   {
      fprintf (stdout, "%%%s-W-SSL, BIO_new(BIO_f_ssl()) failed\n", Utility);
      SeSoLaFree (rqptr);

      if (rqptr->AdminWatchItem &&
          (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA, "BEGIN failed");

      return;
   }

   SSL_set_bio (sesolaptr->SslPtr, sesolaptr->BioPtr, sesolaptr->BioPtr);

   SSL_set_accept_state (sesolaptr->SslPtr);

   if (rqptr->AdminWatchItem &&
       (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
   {
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA, "BEGIN");
      /* set the request pointer as the first item of SSL user data */
      SSL_set_ex_data (sesolaptr->SslPtr, 0, rqptr);
      /* set for SSL information and verification callback */
      SSL_CTX_set_info_callback (sesolaptr->SslPtr, &SesolaWatchInfoCallback);
      /* set for BIO callback with the request pointer as it's argurment */
      BIO_set_callback (sesolaptr->BioPtr, &SesolaWatchBioCallback);
      BIO_set_callback_arg (sesolaptr->BioPtr, rqptr);
   }

   /* begin the SSL handshake */
   SeSoLaAccept (rqptr);
}

/*****************************************************************************/
/*
Shutdown the SSL session with the client (via an SSL alert that we don't wait
around for a response on!)  Return true if RequestEnd() can continue
the request run-down, false if has to abort and wait for some more to happen!
*/

SeSoLaEnd (struct RequestStruct *rqptr)

{
   register struct SeSoLaStruct  *sesolaptr;

   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaEnd() %d\n", rqptr);

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   /* "quiet" shutdown, see [.APPS]S_SERVER.C */
   SSL_set_shutdown (sesolaptr->SslPtr,
                     SSL_SENT_SHUTDOWN | SSL_RECEIVED_SHUTDOWN);

   if (rqptr->AdminWatchItem &&
       (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA, "END");

   SeSoLaFree (rqptr);

   RequestEnd (rqptr);
}

/*****************************************************************************/
/*
Free the OpenSSL structures outside of any use of them (early mistake)!
*/

boolean SeSoLaFree (struct RequestStruct *rqptr)

{
   register struct SeSoLaStruct  *sesolaptr;

   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaFree() %d\n", rqptr);

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   if (sesolaptr->BioPtr != NULL) BIO_free (sesolaptr->BioPtr);

   if (sesolaptr->SslPtr != NULL) SSL_free (sesolaptr->SslPtr);

   /* request has finished with the SESOLA structure */
   VmFree (sesolaptr);
   rqptr->SeSoLaPtr = NULL;
}

/*****************************************************************************/
/*
This establishes the SSL transaction by providing the server "hello",
certificate and key exchange, etc. Due to the non-blocking I/O used by WASD
this function will be called multiple times to complete the OpenSSL accept. 
*/

SeSoLaAccept (struct RequestStruct *rqptr)

{
   register struct SeSoLaStruct  *sesolaptr;

   int  status,
        value;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaAccept() %d\n", rqptr);

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   sesolaptr->ReadType = sesolaptr->WriteType = SESOLA_ACCEPT;

   if (++sesolaptr->AcceptCount == 2)
   {
      /* just finished the first lot of non-blocking I/O */
      if (!memcmp (sesolaptr->AcceptReadDataPtr, "GET ", 4) ||
          !memcmp (sesolaptr->AcceptReadDataPtr, "HEAD ", 5) ||
          !memcmp (sesolaptr->AcceptReadDataPtr, "POST ", 5) ||
          !memcmp (sesolaptr->AcceptReadDataPtr, "PUT ", 4) ||
          !memcmp (sesolaptr->AcceptReadDataPtr, "DELETE ", 7))
      {
         /* looks like it's a standard HTTP request on the wrong port! */
         rqptr->ResponseStatusCode = 400;
         ErrorGeneral (rqptr, ErrorSslPort, FI_LI);
         SeSoLaFree (rqptr);
         RequestEnd (rqptr);
         return;
      }
   }

   value = SSL_accept (sesolaptr->SslPtr);
   if (Debug) fprintf (stdout, "SSL_accept() %d\n", value);

   /* if non-blocking IO in progress just return and wait for delivery */
   if (sesolaptr->ReadInProgress || sesolaptr->WriteInProgress) return;

   if (value > 0)
   {
      /***********/
      /* success */
      /***********/

      if (rqptr->AdminWatchItem &&
          (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
         SesolaWatchSession (rqptr);

      rqptr->NetReadBufferPtr = VmGetHeap (rqptr, NetReadBufferSize);
      rqptr->NetReadBufferSize = NetReadBufferSize;

      NetRead (rqptr, &RequestGet,
               rqptr->NetReadBufferPtr, rqptr->NetReadBufferSize);

      return;
   } 

   /*********/
   /* error */
   /*********/

   if (Debug) ERR_print_errors_fp (stdout);

   /* error, end of request */
   SeSoLaFree (rqptr);
   RequestEnd (rqptr);
}

/*****************************************************************************/
/*
Read application data. This is the equivalent of the general NetRead() function
(and is indeed called from it) in that it reads data from the client via the
network. This function uses SSL_read() to read a stream of encrypted data from
the network then provide it decrypted in the supplied buffer.  This function
provides BLOCKING (non-AST) and NON-BLOCKING I/O conforming to the
functionality provided by NetRead(). The network read I/O status block status
and count fields are valid for the high-level calling routine.
*/

SeSoLaRead
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataSize
)
{
   register struct SeSoLaStruct  *sesolaptr;

   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaRead() %d %d\n", rqptr, AstFunctionPtr);

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   if ((sesolaptr->ReadAstFunctionPtr = AstFunctionPtr) == NULL)
      sesolaptr->ReadType = SESOLA_DATA_BLOCKING;
   else
      sesolaptr->ReadType = SESOLA_DATA_NON_BLOCKING;

   value = SSL_read (sesolaptr->SslPtr,
                     sesolaptr->ReadDataPtr = DataPtr,
                     sesolaptr->ReadDataSize = DataSize);
   if (Debug) fprintf (stdout, "SSL_read() %d\n", value);

   /* if non-blocking IO in progress just return and wait for delivery */
   if (sesolaptr->ReadInProgress) return (SS$_NORMAL);

   /* set the I/O status block count to reflect the decrypted data count */
   if (VMSok (rqptr->NetReadIOsb.Status))
   {
      /* zero returned if received SSL shutdown */
      if (value > 0)
         rqptr->NetReadIOsb.Count = value;
      else
      {
         rqptr->NetReadIOsb.Status = SS$_ABORT;
         rqptr->NetReadIOsb.Count = 0;
         /* let it believe it's now shutdown!  see [.APPS]S_SERVER.C */
         SSL_set_shutdown (sesolaptr->SslPtr,
                           SSL_SENT_SHUTDOWN | SSL_RECEIVED_SHUTDOWN);
      }
   }
   else
   {
      rqptr->NetReadIOsb.Count = DataSize;
      /* let it believe it's now shutdown!  see [.APPS]S_SERVER.C */
      SSL_set_shutdown (sesolaptr->SslPtr,
                        SSL_SENT_SHUTDOWN | SSL_RECEIVED_SHUTDOWN);
   }

   if (AstFunctionPtr == NULL) return (rqptr->NetReadIOsb.Status);

   SysDclAst (AstFunctionPtr, rqptr);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Write application data. This is the equivalent of the general NetWrite()
function (and is called from it) in that it writes data to the client via the
network. This function uses SSL_write() to write non-encrypted application data
to the network in encrypted form. This function provides BLOCKING (non-AST)
and NON-BLOCKING I/O conforming to the functionality provided by NetWrite().
The network write I/O status block status and count fields are valid for the
high-level calling routine.
*/

int SeSoLaWrite
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataLength
)
{
   register struct SeSoLaStruct  *sesolaptr;

   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SeSoLaWrite() %d %d %d %d\n",
               rqptr, AstFunctionPtr, DataPtr, DataLength);
   /** fprintf (stdout, "|%*.*s|\n", DataLength, DataLength, DataPtr); **/

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   if ((sesolaptr->WriteAstFunctionPtr = AstFunctionPtr) == NULL)
      sesolaptr->WriteType = SESOLA_DATA_BLOCKING;
   else
      sesolaptr->WriteType = SESOLA_DATA_NON_BLOCKING;

   value = SSL_write (sesolaptr->SslPtr,
                      sesolaptr->WriteDataPtr = DataPtr,
                      sesolaptr->WriteDataLength = DataLength);
   if (Debug) fprintf (stdout, "SSL_write() %d\n", value);

   /* if non-blocking IO in progress just return and wait for delivery */
   if (sesolaptr->WriteInProgress) return (SS$_NORMAL);

   /* set the I/O status block count to reflect the decrypted data count */
   if (VMSok (rqptr->NetWriteIOsb.Status))
      rqptr->NetWriteIOsb.Count = value;
   else
   {
      rqptr->NetWriteIOsb.Count = 0;
      /* let it believe it's now shutdown!  see [.APPS]S_SERVER.C */
      SSL_set_shutdown (sesolaptr->SslPtr,
                        SSL_SENT_SHUTDOWN | SSL_RECEIVED_SHUTDOWN);
   }

   if (AstFunctionPtr == NULL) return (rqptr->NetWriteIOsb.Status);

   SysDclAst (AstFunctionPtr, rqptr);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Provides OpenSSL status information at various states of SSL processing via the
WATCH facility.
*/

SesolaWatchInfoCallback
(
SSL *SslPtr,
int WhereExactly,
int Value
)
{
   int  WhereGenerally;
   char  *cptr;
   struct  RequestStruct  *rqptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SesolaWatchInfoCallback()\n");

   rqptr = (struct RequestStruct*)SSL_get_ex_data (SslPtr, 0);

   WhereGenerally = WhereExactly & ~SSL_ST_MASK;

   if (WhereGenerally & SSL_ST_CONNECT)
      cptr = "SSL_CONNECT";
   else
   if (WhereGenerally & SSL_ST_ACCEPT)
      cptr = "SSL_ACCEPT";
   else
      cptr = "UNDEFINED";

   if (WhereExactly & SSL_CB_LOOP)
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                      "!AZ !AZ",
                      cptr, SSL_state_string_long(SslPtr));
   else
   if (WhereExactly & SSL_CB_ALERT)
   {
      cptr = (WhereExactly & SSL_CB_READ) ? "read" : "write";
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                      "!AZ SSL3 alert !AZ:!AZ",
                      cptr, SSL_alert_type_string_long(Value),
                      SSL_alert_desc_string_long(Value));
   }
   else
   if (WhereExactly & SSL_CB_EXIT)
   {
      if (Value == 0)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                         "!AZ SSL3 alert failed in !AZ",
                         cptr, SSL_state_string_long(SslPtr));
      else
      if (Value < 0)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                         "!AZ error in !AZ",
                         cptr, SSL_state_string_long(SslPtr));
   }
}

/*****************************************************************************/
/*
Provides OpenSSL Input/Output information each time an SSL read or write occurs
via the WATCH facility.
*/

int SesolaWatchBioCallback
(
BIO *bioptr,
int cmd,
char *argp,
int argi,
long argl,
long retval
)
{
   struct  RequestStruct  *rqptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SesolaWatchBioCallback()\n");

   rqptr = (struct RequestStruct*) BIO_get_callback_arg (bioptr);

   if (cmd == (BIO_CB_READ | BIO_CB_RETURN))
   {
      if (retval == -1)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                         "READ !UL bytes non-blocking", argi);
      else
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                         "READ !UL bytes done", retval);
      return (retval);
   }
   else
   if (cmd == (BIO_CB_WRITE | BIO_CB_RETURN))
   {
      if (retval == -1)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                         "WRITE !UL bytes non-blocking", argi);
      else
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                         "WRITE !UL bytes done", retval);
      return (retval);
   }

   return (retval);
}

/*****************************************************************************/
/*
Provide WATCH information after final SSL session establishment show version
and cipher in use.
*/

SeSoLaWatchSession (struct RequestStruct *rqptr)

{
   int  KeySize;
   char  *CipherNamePtr,
         *VersionNamePtr;
   SSL  *SslPtr;
   SSL_CIPHER  *CipherPtr;
   SSL_SESSION  *SessionPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaWatchSession()\n");

   SslPtr = rqptr->SeSoLaPtr->SslPtr;
   SessionPtr = SSL_get_session (SslPtr);
   CipherPtr = SSL_get_current_cipher (SslPtr);

   CipherNamePtr = SSL_CIPHER_get_name (CipherPtr);
   KeySize = SSL_CIPHER_get_bits (CipherPtr, NULL);
   VersionNamePtr = SSL_get_version (SslPtr);

   AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                   "SESSION cached:!AZ protocol:!AZ cipher:!AZ keysize:!UL",
                   SslPtr->hit ? "yes" : "no",
                   VersionNamePtr, CipherNamePtr, KeySize);
}

/*****************************************************************************/
/*
Generate some SSL-related CGI variables.  Called from CgiGenerateVariables().
*/ 

int SesolaCgiGenerateVariables
(
struct RequestStruct *rqptr,
int VarType
)
{
   static $DESCRIPTOR (NumberFaoDsc, "!UL\0");

   register char  *cptr;

   int  status,
        KeySize;
   char  String [512];
   $DESCRIPTOR (StringDsc, String);
   SSL  *SslPtr;
   SSL_CIPHER  *CipherPtr;
   SSL_SESSION  *SessionPtr;
   X509  *ServerCertPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SesolaCgiGenerateVariables()\n");

   if (rqptr->SeSoLaPtr == NULL)
   {
      if (VMSnok (status =
          CgiVariable (rqptr, "SECURITY_STATUS", "NONE", VarType)))
         return (status);

      return (SS$_NORMAL);
   }

   SslPtr = rqptr->SeSoLaPtr->SslPtr;
   SessionPtr = SSL_get_session (SslPtr);
   CipherPtr = SSL_get_current_cipher (SslPtr);
   ServerCertPtr = SSL_get_certificate (SslPtr);

   cptr = SSL_CIPHER_get_name (CipherPtr);
   if (VMSnok (status =
       CgiVariable (rqptr, "SSL_CIPHER", cptr, VarType)))
      return (status);

   KeySize = SSL_CIPHER_get_bits (CipherPtr, NULL);
   sys$fao (&NumberFaoDsc, 0, &StringDsc, KeySize);
   if (VMSnok (status =
       CgiVariable (rqptr, "SSL_CIPHER_KEYSIZE", String, VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SSL_CLIENT_DN", "NONE", VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SSL_CLIENT_CA", "NONE", VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SSL_CLIENT_AUTHENTICATED", "FALSE", VarType)))
      return (status);

   if (VMSnok (status =
       CgiVariable (rqptr, "SECURITY_STATUS", "SSL", VarType)))
      return (status);

   X509_NAME_oneline (X509_get_subject_name(ServerCertPtr),
                      String, sizeof(String));
   if (VMSnok (status =
       CgiVariable (rqptr, "SSL_SERVER_DN", String, VarType)))
      return (status);

   X509_NAME_oneline (X509_get_issuer_name(ServerCertPtr),
                      String, sizeof(String));
   if (VMSnok (status =
       CgiVariable (rqptr, "SSL_SERVER_CA", String, VarType)))
      return (status);

   cptr = SSL_get_version (SslPtr);
   if (VMSnok (status =
       CgiVariable (rqptr, "SSL_VERSION", cptr, VarType)))
      return (status);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Brief report on this port's SSL context statistics and session information.
*/

SeSoLaReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... SSL Report</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>SSL Report</H2>\n\
!AZ\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>!AZ</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Version:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Build:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Protocol:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Certificate:</TH><TD>!AZ</TD></TR>\n\
<TR><TH></TH><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=2 BORDER=0>\n\
<TR><TH ALIGN=right VALIGN=top>Issuer:&nbsp;</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right VALIGN=top>Server:&nbsp;</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Expires:&nbsp;</TH><TD>!AZ</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Private Key:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Cipher List:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right VALIGN=top>Supported Ciphers:</TH><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TH></TH><TH ALIGN=left>Version</TH><TH ALIGN=left>Name</TH></TR>\n");

   static $DESCRIPTOR (CiphersFaoDsc,
"<TR><TH ALIGN=right>!UL. &nbsp;</TH>\
<TD><TT>!AZ</TT> &nbsp;</TD>\
<TD><TT>!AZ</TT> &nbsp;</TD>\
</TR>\n");

   static $DESCRIPTOR (CurrentSession1FaoDsc,
"</TABLE>\n\
</TD></TR>\n\
<TR><TH ALIGN=right VALIGN=top>Service Accepts:</TH><TD VALIGN=top>\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>Total:</TH><TD ALIGN=left> &nbsp;!UL</TD></TR>\n\
<TR><TH ALIGN=right> &nbsp;Completed:</TH><TD ALIGN=left> &nbsp;!UL</TD></TR>\n\
</TABLE>\n\
<TR><TH ALIGN=right VALIGN=top>Session Cache:</TH><TD VALIGN=top>\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>Hits:</TH><TD ALIGN=left> &nbsp;!UL</TD></TR>\n\
<TR><TH ALIGN=right>Misses:</TH><TD ALIGN=left> &nbsp;!UL</TD></TR>\n\
<TR><TH ALIGN=right> &nbsp;Timeouts:</TH><TD ALIGN=left> &nbsp;!UL</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Current Session</TH></TR>\n\
<TR><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Cache Hit:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Version:</TH><TD><TT>!AZ</TD></TT></TR>\n\
<TR><TH ALIGN=right VALIGN=top>Shared Ciphers:</TH><TD>");

   static $DESCRIPTOR (SharedCipherFaoDsc,
"!AZ &nbsp;<B>!UL.</B> &nbsp;<TT>!AZ</TT>\n");

   static $DESCRIPTOR (CurrentSession2FaoDsc,
"</TD></TR>\n\
<TR><TH ALIGN=right>Session Cipher:</TH><TD><TT>!AZ</TT></TD></TR>\n\
<TR><TH ALIGN=right>Key Size:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
</BODY>\n\
</HTML>\n");

   register unsigned long  *vecptr;
   register char  *cptr, *sptr, *zptr;
   register struct SeSoLaStruct  *sesolaptr;

   int  status,
        Count,
        Total;
   unsigned short  Length;
   unsigned long  BinaryTime [2],
                  /* WARNING: really does need to be fairly large!! */
                  FaoVector [256],
                  ResultTime [2];
   char  Buffer [2048],
         CertString [256],
         CertCaString [256],
         CertDnString [256],
         CertExpiresString [64],
         SharedCiphers [256];
   $DESCRIPTOR (BufferDsc, Buffer);
   STACK  *StackPtr;
   SSL  *SslPtr;
   SSL_CIPHER  *CipherPtr;
   SSL_SESSION  *SessionPtr;
   X509  *ServerCertPtr;
   X509  *PeerPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSolaReport()\n");

   if (rqptr->RequestScheme != SCHEME_HTTPS)
   {
      ErrorGeneral (rqptr, ErrorSslReportService, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   SslPtr = rqptr->SeSoLaPtr->SslPtr;
   ServerCertPtr = SSL_get_certificate (SslPtr);

   X509_NAME_oneline (X509_get_issuer_name(ServerCertPtr),
                      CertString, sizeof(CertString));
   SesolaReportFormatCertName (CertString, CertCaString, sizeof(CertCaString));

   X509_NAME_oneline (X509_get_subject_name(ServerCertPtr),
                      CertString, sizeof(CertString));
   SesolaReportFormatCertName (CertString, CertDnString, sizeof(CertDnString));

   ASN1_TIME_print (SesolaBioMemPtr, X509_get_notAfter(ServerCertPtr));
   BIO_gets (SesolaBioMemPtr, CertExpiresString, sizeof(CertExpiresString));

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   *vecptr++ = rqptr->ServicePtr->ServerHostPort;
   *vecptr++ = SSLeay_version (SSLEAY_VERSION);
   cptr = SSLeay_version (SSLEAY_BUILT_ON);
   if ((sptr = strstr (cptr, "uilt on")) != NULL)
   {
      cptr = sptr + 7;
      while (*cptr && (*cptr == ' ' || *cptr == ':')) cptr++;
   }
   *vecptr++ = cptr;

   *vecptr++ = SeSoLaProtocolsPtr;
   *vecptr++ = rqptr->ServicePtr->SeSoLaCert;
   *vecptr++ = CertCaString;
   *vecptr++ = CertDnString;
   *vecptr++ = CertExpiresString;
   *vecptr++ = rqptr->ServicePtr->SeSoLaKey;
   if (rqptr->ServicePtr->SeSoLaCipherList[0])
      *vecptr++ = rqptr->ServicePtr->SeSoLaCipherList;
   else
      *vecptr++ = "<I>(none)</I>";

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   if (rqptr->RequestScheme != SCHEME_HTTPS)
   {
      NetWriteBuffered (rqptr, NextTaskFunction, "</BODY>\n</HTML>\n", 16);
      return;
   }

   /*********************/
   /* supported ciphers */
   /*********************/

   StackPtr = SSL_get_ciphers (sesolaptr->SslPtr);
   Total = sk_num (StackPtr);
   for (Count = 0; Count < Total; Count++)
   {
      CipherPtr = (SSL_CIPHER *)sk_value (StackPtr, Count);

      vecptr = FaoVector;
      *vecptr++ = Count + 1;
      *vecptr++ = SSL_CIPHER_get_version (CipherPtr);
      *vecptr++ = SSL_CIPHER_get_name (CipherPtr);

      status = sys$faol (&CiphersFaoDsc, &Length, &BufferDsc, &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, FI_LI);
         SysDclAst (NextTaskFunction, rqptr);
         return;
      }
      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   vecptr = FaoVector;

   *vecptr++ = SSL_CTX_sess_accept (sesolaptr->SeSoLaCtx);
   *vecptr++ = SSL_CTX_sess_accept_good (sesolaptr->SeSoLaCtx);
   *vecptr++ = SSL_CTX_sess_hits (sesolaptr->SeSoLaCtx);
   *vecptr++ = SSL_CTX_sess_misses (sesolaptr->SeSoLaCtx);
   *vecptr++ = SSL_CTX_sess_timeouts (sesolaptr->SeSoLaCtx);

   /*******************/
   /* current session */
   /*******************/

   if (sesolaptr->SslPtr->hit)
      *vecptr++ = "yes";
   else
      *vecptr++ = "no";

   CipherPtr = SSL_get_current_cipher (sesolaptr->SslPtr);

   /* can't really be null or we would not be here! */
   if (CipherPtr != NULL)
      *vecptr++ = SSL_CIPHER_get_version (CipherPtr);

   cptr = sptr = SSL_get_shared_ciphers (sesolaptr->SslPtr,
                    SharedCiphers, sizeof(SharedCiphers));
   Count = 0;
   /* there has to be at least one or we would not be here! */
   if (cptr != NULL)
   {
      while (*cptr)
      {
         while (*cptr && *cptr != ':') cptr++;
         if (*cptr) cptr++;
         Count++;
      }
   }
   *vecptr++ = Count;

   status = sys$faol (&CurrentSession1FaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   /* has to be at least one or we would not be here! */
   cptr = sptr;
   if (cptr != NULL)
   {
      Count = 0;
      while (*cptr)
      {
         vecptr = FaoVector;

         sptr = cptr;
         while (*cptr && *cptr != ':') cptr++;
         if (*cptr) *cptr++ = '\0';

         if (Count++)
            *vecptr++ = "<BR>";
         else
            *vecptr++ = "";

         *vecptr++ = Count;
         *vecptr++ = sptr;

         status = sys$faol (&SharedCipherFaoDsc, &Length, &BufferDsc,
                            &FaoVector);
         if (VMSnok (status) || status == SS$_BUFFEROVF)
         {
            rqptr->ErrorTextPtr = "sys$faol()";
            ErrorVmsStatus (rqptr, status, FI_LI);
            SysDclAst (NextTaskFunction, rqptr);
            return;
         }
         Buffer[Length] = '\0';
         NetWriteBuffered (rqptr, 0, Buffer, Length);
      }
   }

   vecptr = FaoVector;

   *vecptr++ = SSL_CIPHER_get_name (CipherPtr);
   *vecptr++ = SSL_CIPHER_get_bits (CipherPtr, NULL);

   status = sys$faol (&CurrentSession2FaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, NextTaskFunction, Buffer, Length);
}

/*****************************************************************************/
/*
Just <BR> the certificate DN format at each component.
*/

SesolaReportFormatCertName
(
char *SourceName,
char *FormatName,
int SizeOfFormatName
)
{
   register char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SesolaReportFormatCertName() |%s|\n", SourceName);

   zptr = (sptr = FormatName) + SizeOfFormatName - 6;
   cptr = SourceName;
   while (*cptr && sptr < zptr)
   {
      *sptr++ = *cptr++;
      if (*cptr == '/' && sptr < zptr)
      {
         *(unsigned long*)sptr = '<BR>';
         sptr += 4;
      }
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", FormatName);
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

BIO_METHOD *BIO_s_sesola()
{
   return (&sesola_method);
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

int sesola_new (BIO *bioptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "sesola_new()\n");

   bioptr->init = 1;
   bioptr->shutdown = 1;
   bioptr->num = 0;
   bioptr->flags = 0;
   /* the BIO pointer is set to the SESOLA structure by the calling routine */
   return (1);
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

int sesola_free (BIO *bioptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "sesola_free()\n");

   if (bioptr == NULL) return (0);
   /* don't free anything, the BIO pointer is set to the SESOLA structure */
   bioptr->ptr = NULL;
   return (1);
}
	
/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

int sesola_read
(
BIO *bioptr,
char *DataPtr,
int DataSize
)
{
   register struct SeSoLaStruct  *sesolaptr;
   register struct RequestStruct  *rqptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "sesola_read() %d %d\n", DataPtr, DataSize);

   sesolaptr = (struct SeSoLaStruct*)bioptr->ptr;
   rqptr = sesolaptr->rqptr;

   if (Debug)
      fprintf (stdout,
         "ReadCompleted: %d ReadInProgress: %d WriteInProgress: %d\n",
         sesolaptr->ReadCompleted, sesolaptr->ReadInProgress,
         sesolaptr->WriteInProgress);

   if (sesolaptr->ReadInProgress)
   {
      /* asynchronous read in progress, retry next call */
      BIO_set_retry_read (bioptr);
      return (-1);
   }

   if (sesolaptr->ReadCompleted)
   {
      /* previous non-blocking read */
      sesolaptr->ReadCompleted = false;
      BIO_clear_retry_flags (bioptr);
      /* in case of an error give it zero (shutdown) */
      if (VMSok (rqptr->NetReadIOsb.Status))
         return (rqptr->NetReadIOsb.Count);
      else
         return (0);
   }

   if (sesolaptr->ReadType == SESOLA_DATA_BLOCKING)
   {
      /* blocking I/O */
      sesolaptr->ReadInProgress = false;
      BIO_clear_retry_flags (bioptr);
      status = NetReadRaw (rqptr, NULL, DataPtr, DataSize);
      /* if success return the number of characters written */
      if (VMSok (status)) return (rqptr->NetReadIOsb.Count);
      /* whoa, error ... will be checked for by SeSoLaRead() */
      rqptr->NetReadIOsb.Status = status;
      return (DataSize);
   }
   else
   {
      /* non-blocking I/O */
      sesolaptr->ReadInProgress = true;
      BIO_set_retry_read (bioptr);
      if (sesolaptr->ReadType == SESOLA_ACCEPT)
         sesolaptr->AcceptReadDataPtr = DataPtr;
      NetReadRaw (rqptr, &sesola_read_ast, DataPtr, DataSize);
      /* return indicating non-blocking I/O */
      return (-1);
   }
}

/*****************************************************************************/
/*
*/

int sesola_read_ast (struct RequestStruct *rqptr)

{
   register struct SeSoLaStruct  *sesolaptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "sesola_read_ast() %%X%08.08X %d\n",
               rqptr->NetReadIOsb.Status, rqptr->NetReadIOsb.Count);

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   sesolaptr->ReadInProgress = false;
   sesolaptr->ReadCompleted = true;

   if (sesolaptr->ReadType == SESOLA_DATA_NON_BLOCKING)
      SeSoLaRead (rqptr, sesolaptr->ReadAstFunctionPtr,
                  sesolaptr->ReadDataPtr, sesolaptr->ReadDataSize);
   else
   if (sesolaptr->ReadType == SESOLA_ACCEPT)
      SeSoLaAccept (rqptr);
   else
   if (sesolaptr->ReadType == SESOLA_SHUTDOWN)
      SeSoLaEnd (rqptr);
   else
      ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

int sesola_write
(
BIO *bioptr,
char *DataPtr,
int DataLength
)
{
   register struct SeSoLaStruct  *sesolaptr;
   register struct RequestStruct  *rqptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "sesola_write() %d %d\n", DataPtr, DataLength);

   sesolaptr = (struct SeSoLaStruct*)bioptr->ptr;
   rqptr = sesolaptr->rqptr;

   if (Debug)
      fprintf (stdout,
         "WriteCompleted: %d ReadInProgress: %d WriteInProgress: %d\n",
         sesolaptr->WriteCompleted, sesolaptr->ReadInProgress,
         sesolaptr->WriteInProgress);

   if (sesolaptr->WriteInProgress)
   {
      /* asynchronous write in progress, retry next call */
      BIO_set_retry_write (bioptr);
      return (-1);
   }

   if (sesolaptr->WriteCompleted)
   {
      /* previous non-blocking write */
      sesolaptr->WriteCompleted = false;
      BIO_clear_retry_flags (bioptr);
      /* if an error give it what it wants, error is detected SeSolaWrite() */
      if (VMSok (rqptr->NetWriteIOsb.Status))
         return (rqptr->NetWriteIOsb.Count);
      else
         return (DataLength);
   }

   if (sesolaptr->WriteType == SESOLA_DATA_BLOCKING)
   {
      /* blocking I/O */
      sesolaptr->WriteInProgress = sesolaptr->WriteCompleted = false;
      BIO_clear_retry_flags (bioptr);
      status = NetWriteRaw (sesolaptr->rqptr, NULL, DataPtr, DataLength);
      /* if success return the number of characters written */
      if (VMSok (status)) return (rqptr->NetWriteIOsb.Count);
      /* whoa, error ... will be checkd for by SeSoLaWrite() */
      rqptr->NetWriteIOsb.Status = status;
      return (DataLength);
   }
   else
   {
      /* non-blocking I/O */
      sesolaptr->WriteInProgress = true;
      /* indicate to the calling routine this write should be retried */
      BIO_set_retry_write (bioptr);
      NetWriteRaw (sesolaptr->rqptr, &sesola_write_ast, DataPtr, DataLength);
      /* return indicating non-blocking I/O */
      return (-1);
   }
}

/*****************************************************************************/
/*
*/

int sesola_write_ast (struct RequestStruct *rqptr)

{
   register struct SeSoLaStruct  *sesolaptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "sesola_write_ast() %%X%08.08X %d\n",
               rqptr->NetWriteIOsb.Status, rqptr->NetWriteIOsb.Count);

   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   sesolaptr->WriteInProgress = false;
   sesolaptr->WriteCompleted = true;

   if (sesolaptr->WriteType == SESOLA_DATA_NON_BLOCKING)
      SeSoLaWrite (rqptr, sesolaptr->WriteAstFunctionPtr,
                   sesolaptr->WriteDataPtr, sesolaptr->WriteDataLength);
   else
   if (sesolaptr->WriteType == SESOLA_ACCEPT)
      SeSoLaAccept (rqptr);
   else
   if (sesolaptr->WriteType == SESOLA_SHUTDOWN)
      SeSoLaEnd (rqptr);
   else
      ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

long sesola_ctrl
(
BIO *bioptr,
int Command,
long Number,
char *Pointer
)
{
   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "sesola_ctrl() %d %d %d\n", Command, Number, Pointer);

   value = 1;
   switch (Command)
   {
      case BIO_CTRL_RESET:
         if (Debug) fprintf (stdout, "BIO_CTRL_RESET\n");
         return (1);

      case BIO_CTRL_EOF:
         if (Debug) fprintf (stdout, "BIO_CTRL_EOF\n");
         return (1);

      case BIO_CTRL_SET:
         if (Debug) fprintf (stdout, "BIO_CTRL_SET\n");
         bioptr->num = Number;
         return (1);

      case BIO_CTRL_SET_CLOSE:
         if (Debug) fprintf (stdout, "BIO_CTRL_SET_CLOSE\n");
         return (1);

      case BIO_CTRL_FLUSH:
         if (Debug) fprintf (stdout, "BIO_CTRL_FLUSH\n");
         return (1);

      case BIO_CTRL_DUP:
         if (Debug) fprintf (stdout, "BIO_CTRL_DUP\n");
         return (1);

      case BIO_CTRL_GET_CLOSE:
         if (Debug) fprintf (stdout, "BIO_CTRL_GET_CLOSE\n");
         return (0);

      case BIO_CTRL_INFO:
         if (Debug) fprintf (stdout, "BIO_CTRL_INFO\n");
         return (0);

      case BIO_CTRL_GET:
         if (Debug) fprintf (stdout, "BIO_CTRL_GET\n");
         return (0);

      case BIO_CTRL_PENDING:
         if (Debug) fprintf (stdout, "BIO_CTRL_PENDING\n");
         return (0);

      case BIO_CTRL_POP:
         if (Debug) fprintf (stdout, "BIO_CTRL_POP\n");
         return (0);

      case BIO_CTRL_PUSH:
         if (Debug) fprintf (stdout, "BIO_CTRL_PUSH\n");
         return (0);

      case BIO_CTRL_WPENDING:
         if (Debug) fprintf (stdout, "BIO_CTRL_WPENDING\n");
         return (0);

      default:
         if (Debug) fprintf (stdout, "default:\n");
         return (0);
   }
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

int sesola_gets
(
BIO *bioptr,
char *StringPtr,
int StringSize
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "sesola_gets()\n");
   return(0);
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

int sesola_puts
(
BIO *bioptr,
char *StringPtr
)
{
    int  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "sesola_puts() |%s|\n", StringPtr);

    if (StringPtr == NULL) return (0);
    Length = strlen (StringPtr);
    if (Length == 0) return (0);
    return (sesola_write (bioptr, StringPtr, Length));
}

/*****************************************************************************/
/*
For compilations without SSL these functions provide LINKage stubs for the
rest of the HTTPd modules, allowing for just recompiling the SESOLA module to
integrate the SSL functionality.
*/

/*********************/
#else  /* not SESOLA */
/*********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

/* global storage */
boolean  ProtocolHttpsAvailable,
         ProtocolHttpsConfigured;
/* i.e. disabled */
int  SeSoLaProtocolVersion = -1;
char  HttpdSesola [] = "",
      SeSoLaParams [256] = "";


/* external storage */
extern char  ErrorSanityCheck[];
extern char  Utility[];

SeSoLaInit ()
{
   ProtocolHttpsAvailable = false;
   if (SeSoLaProtocolVersion <= 0) return;
   fprintf (stdout, "%%%s-E-SSL, non-SSL version\n", Utility);
   exit (STS$K_ERROR | STS$M_INHIB_MSG);
}

SeSoLaInitServiceList ()
{
   return;
}

SeSoLaBegin (struct RequestStruct *rqptr)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SeSoLaAccept (struct RequestStruct *rqptr)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SeSoLaEnd (struct RequestStruct *rqptr)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SeSoLaFree (struct RequestStruct *rqptr)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SeSoLaShutdown (struct RequestStruct *rqptr)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SeSoLaWrite
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataLength
)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SeSoLaRead
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataSize
)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SesolaCgiGenerateVariables
(
struct RequestStruct *rqptr,
int VarType
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SesolaCgiGenerateVariables()\n");

   if (VMSnok (status =
       CgiVariable (rqptr, "SECURITY_STATUS", "NONE", VarType)))
      return (status);

   return (SS$_NORMAL);
}

SeSoLaReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   ErrorGeneral (rqptr, "This is a non-SSL version.", FI_LI);
   SysDclAst (NextTaskFunction, rqptr);
}

/************************/
#endif  /* ifdef SESOLA */
/************************/

/*****************************************************************************/

